import IPathAnnotation from './IPathAnnotation';

/** 
 * @memberof Core.Annotations
 */
class FreeTextAnnotation extends IPathAnnotation {

  /**
   * An enum with possible auto-size types.
   * @property {string} NONE - Freetext stays at fixed size
   * @property {string} AUTO - Autosize to content
   * @property {string} FIXED_WIDTH - Width cannot be changed
   * @property {string} FIXED_HEIGHT - Height cannot be changed
   */
  static AutoSizeTypes = {
    NONE: 'NONE',
    AUTO: 'AUTO',
    FIXED_WIDTH: 'FIXED_WIDTH',
    FIXED_HEIGHT: 'FIXED_HEIGHT'
  };

  /**
   * An enum with possible freetext intents.
   * @property {string} FreeText - Treat as regular freetext
   * @property {string} FreeTextCallout - Treat as callout
   */
  static Intent = {
    FreeText: 'FreeText',
    FreeTextCallout: 'FreeTextCallout'
  };

  /**
   * Represents a free text annotation.
   * @extents Core.Annotations.IPathAnnotation
   * 
   * @param {string} [intent] - The intended usage type of this freetext annotation. Could either be freetext or callout.
   * @param {object} [initializer] - A map of values to auto-initialize certain properties of the annotation. You can only initialize properties defined on the annotation under the Members section (unless specified otherwise).
   * 
   * @property {boolean} EXPORT_CALCULATED_FONT_SIZE - Whether to export calculated font size for auto-sized fonts. Default: true
   * @property {object} Intent - An enumeration of possible intents for FreeText Annotations
   * @property {string} Intent.FreeText - Should treat as regular freetext
   * @property {string} Intent.FreeTextCallout - Should treat as callout
   */
  constructor(intent = null, initializer = {}) {
    super(initializer);
    this.EXPORT_CALCULATED_FONT_SIZE = true;
    this.Intent = intent || FreeTextAnnotation.Intent.FreeText;

    Object.keys(initializer).forEach(key => {
      if (this.hasOwnProperty(key)) {
        this[key] = initializer[key];
      }
    });

    /**
     * Object that contains all the flags to customize the auto-sizing behavior
     * @property {boolean} shrinkWidth - Whether to shrink the width of the annotation if the size of its contents are smaller in width than the annotation
     * @property {boolean} shrinkHeight - Whether to shrink the height of the annotation if the size of its contents are smaller in height than the annotation
     * @property {boolean} expandWidth - Whether to expand the width of the annotation if the size of the contents are larger in width than the annotation
     * @property {boolean} expandHeight - Whether to expand the height of the annotation if the size of the contents are larger in height than the annotation
     */
    this.autoSizeProperties = {
      shrinkWidth: false,
      shrinkHeight: true,
      expandWidth: false,
      expandHeight: true
    };

    /** The font to use for the annotation's text. */
    this.Font = '';

    /** The font size to use for the annotation's text, specified like a CSS font size. */
    this.FontSize = 14;

    /** The horizontal alignment of the annotation's text (left, right, center) */
    this.TextAlign = 'left';

    /** The color of the text in the annotation. */
    this.TextColor = '';

    /** The vertical alignment of the annotation's text (top, bottom, center) default: top */
    this.TextVerticalAlign = 'top';
  }

  /** Disables the Enter key for the instance of RichTextEditor */
  disableEnterKeypress() {}

  /** Enables the Enter key for the instance of RichTextEditor */
  enableEnterKeypress() {}

  /**
   * Resize the annotation's text box to fit its contents
   * @param {Core.Document.PageInfo} pageInfo - An object representing the page info. Contains the properties "width" and "height".
   * @param {object} pageMatrix - The transformation matrix for the page that the annotation is on.
   * @param {number} pageRotation - The internal degrees of rotation of the current page.
   * @example
const freetextAnnot = annotManager.getAnnotationsList()[0];
const doc = docViewer.getDocument();
const pageNumber = 1;
const pageInfo = doc.getPageInfo(pageNumber);
const pageMatrix = doc.getPageMatrix(pageNumber);
const pageRotation = doc.getPageRotation(pageNumber);
freetextAnnot.fitText(pageInfo, pageMatrix, pageRotation);
annotManager.drawAnnotationsFromList([freetextAnnot]);
   */
  fitText(pageInfo, pageMatrix, pageRotation) {}

  /**
   * Gets the auto-sizing method if any.
   * @return {string} The type of autosizing method
   */
  getAutoSizeType() {}

  /**
   * Gets the calculated font size when auto-sizing the text font size to fit the free text annotation's bounding box. This method retrieves the font size that has been dynamically calculated to fit the content within the bounding box of the free text annotation when Core.Annotations.FreeTextAnnotation#switchToAutoFontSize is used to enable auto-sizing of the font size. If Core.Annotations.FreeTextAnnotation#switchToAutoFontSize has not been called, it will return the current font size.
   * @return {string} The calculated font size of the annotation, including the unit (e.g., "16px").
   * @example
WebViewer(...)
.then(instance => {
  const { annotationManager, Annotations } = instance.Core;

  annotationManager.addEventListener('annotationSelected', (annotations, action) => {
    if(action === 'selected' && annotations[0] instanceof Annotations.FreeTextAnnotation) {
      annotations[0].switchToAutoFontSize();
    }
  });

  annotationManager.addEventListener('annotationChanged', (annotations, action) => {
    if(action === 'modify' && annotations[0] instanceof Annotations.FreeTextAnnotation) {
      console.log(annotations[0].getCalculatedFontSize())
    }
  });
});
   */
  getCalculatedFontSize() {}

  /**
   * Gets the date format of the FreeText (if any)
   * @return {string} Returns Date format
   */
  getDateFormat() {}

  /** @return {Core.Annotations.FreeTextAnnotation.RichTextEditor} Returns the rich text editor instance associated with the annotation */
  getEditor() {}

  /**
   * Gets the intent of the FreeText annotation e.g. FreeTextCallout
   * @return {string} The intent of the annotation
   */
  getIntent() {
    return this.Intent;
  }

  /**
   * Get the rectangle difference of the annotation bounding rect and the text rect
   * @return {Core.Math.Rect} the rectangle difference
   */
  getPadding() {}

  /**
   * Gets the start style of the line. Only applicable for callout annotations
   * @return {string} The start style
   */
  getStartStyle() {}

  /** @return {boolean} Returns a boolean which indicates if the bounding box of the freetext will be resized automatically when its contents are being edited */
  isAutoSized() {}

  /** @return {boolean} Returns whether the current font is Automatically sized */
  isAutoSizeFont() {}

  /**
   * Sets how the freetext annotation auto-sizes to content. If you are passing 'AUTO' as the parameter, you will need to invoke the Core.Annotations.FreeTextAnnotation#fitText method afterwards.
   * @param {string} type - Use one of the types provided by AutoSizeTypes
   */
  setAutoSizeType(type) {}

  /**
   * Set FreeText as a date type annotation with the specified format. A viewer can use this information to allow easier updating of the text to represent a date, for example by using a date picker.
   * @param {string} format - Set Annotation Date format
   */
  setDateFormat(format) {}

  /**
   * Sets the intent of the FreeText annotation
   * @param {string} intent - The new intent for the annotation
   */
  setIntent(intent) {
    this.Intent = intent;
  }

  /**
   * Sets the rectangle difference of the annotation bounding rect and the text rect
   * @param {Core.Math.Rect} rect - The new rectangle difference
   */
  setPadding(rect) {}

  /**
   * Sets the start style of the line. Only applicable for callout annotations.
   * @param {string} startStyle - The starting style
   */
  setStartStyle(startStyle) {}

  /**
   * Sets the font size of the free text annotation to automatically adjust to fit its bounding box when resized. This method is used to switch the font size behavior of the free text annotation to "auto," which means that the font size will adapt dynamically to the size of the annotation's bounding box when it is resized.
   * @example
WebViewer(...)
.then(instance => {
  const { documentViewer, annotationManager, Annotations } = instance.Core;

  documentViewer.addEventListener('annotationsLoaded', () => {
    const annotList = annotationManager.getAnnotationsList();
    annotList.forEach(annot => {
      if (annot instanceof Annotations.FreeTextAnnotation) {
        annot.switchToAutoFontSize();
      }
    });
  });
});
   */
  switchToAutoFontSize() {}
}