/**
 * @memberof Core.Annotations
 */
class WidgetFlags {
  /** If set, the field shall be automatically divided into as many equally spaced positions as the value of the max length. */
  static COMB

  /** If set, the field is a combo box; if clear, the field is a list box. */
  static COMBO

  /** If set, the new value shall be committed as soon as a selection is made. */
  static COMMIT_ON_SEL_CHANGE

  /** If set, the field shall not scroll (horizontally for single-line fields, vertically for multiple-line fields) to accommodate more text than fits within its annotation rectangle. */
  static DO_NOT_SCROLL

  /** If set, text entered in the field shall not be spell-checked. */
  static DO_NOT_SPELL_CHECK

  /** If set, the combo box shall include an editable text box as well as a drop-down list; if clear, it shall include only a drop-down list. */
  static EDIT

  /** If set, the text entered in the field represents the pathname of a file whose contents shall be submitted as the value of the field. */
  static FILE_SELECT

  /** If set, more than one of the field’s option items may be selected simultaneously; if clear, at most one item shall be selected. */
  static MULTI_SELECT

  /** If set, the field may contain multiple lines of text. */
  static MULTILINE

  /** If set, the field shall not be exported by a submit-form action. */
  static NO_EXPORT

  /** (Radio buttons only) If set, exactly one radio button shall be selected at all times. */
  static NO_TOGGLE_TO_OFF

  /** If set, the field is intended for entering a secure password that should not be echoed visibly to the screen. */
  static PASSWORD

  /** If set, the field is a pushbutton that does not retain a permanent value. */
  static PUSH_BUTTON

  /** If set, the field is a set of radio buttons; if clear, the field is a checkbox. */
  static RADIO

  /** If set, a group of radio buttons within a radio button field that use the same value for the on state will turn on and off in unison. */
  static RADIOS_IN_UNISON

  /** If set, the user may not change the value of the field. */
  static READ_ONLY

  /** If set, the field shall have a value at the time it is exported by a submit-form action. */
  static REQUIRED

  /** If set, the value of this field shall be a rich text string. */
  static RICH_TEXT

  /** If set, the field’s option items shall be sorted alphabetically. */
  static SORT

  /**
   * An object that describes flags on a Widget, or other PDF objects.
   * @param {?Object} [options] - Define the set of supported flags.
   */
  constructor(options = {}) {
    const self = this;
    Object.keys(options).forEach(key => {
      if (key in self) self[key] = options[key];
    });
  }

  /**
   * Determine if the given flag is set.
   * @param {string} flag The name of the flag to check
   * @return {boolean} Whether the flag is set
   */
  get(flag) {
    return !!this[flag];
  }

  /**
   * Set flags to true or false. If options is an array of strings, set the flags with those names to true. If options is a number or array of numbers, set the flags from the corresponding numbers to true. (deprecated, relies on a limited interpretation of numbers as bitfields) If options is a string, set that flag to shouldSet.
   * @param {Array.<string>|Array.<number>|string|number} options - One of the possible input types to determine which flags to set
   * @param {boolean} [shouldSet] - If options is a string, whether to set or unset the flag
   */
  set(options, shouldSet = true) {
    if (Array.isArray(options)) {
      options.forEach(option => {
        if (typeof option === 'string') {
          this.flags[option] = true;
        } else if (typeof option === 'number') {
          const flagName = `flag${option}`;
          this.flags[flagName] = true;
        }
      });
    } else if (typeof options === 'string') {
      this.flags[options] = shouldSet;
    } else if (typeof options === 'number') {
      const flagName = `flag${options}`;
      this.flags[flagName] = true;
    }
  }
}