/** 
 * The namespace for anything to do with PDF annotations.
 * @memberof Core
 * @namespace
 */
const Annotations = {
  /**
   * An enum representing different line end types that are available for line annotations
   * @property {string} NONE - No line endings
   * @property {string} OPEN_ARROW - an arrow that points outward
   * @property {string} R_OPEN_ARROW - an arrow that points inward
   * @property {string} CLOSED_ARROW - a triangle that points outward
   * @property {string} R_CLOSED_ARROW - a triangle that points inward
   * @property {string} BUTT - a vertical line
   * @property {string} SQUARE - a square
   * @property {string} DIAMOND - a diamond
   * @property {string} CIRCLE - a circle
   * @property {string} SLASH - a slash
   */
  LineEndType: {
    NONE: 'NONE',
    OPEN_ARROW: 'OPEN_ARROW',
    R_OPEN_ARROW: 'R_OPEN_ARROW',
    CLOSED_ARROW: 'CLOSED_ARROW',
    R_CLOSED_ARROW: 'R_CLOSED_ARROW',
    BUTT: 'BUTT',
    SQUARE: 'SQUARE',
    DIAMOND: 'DIAMOND',
    CIRCLE: 'CIRCLE',
    SLASH: 'SLASH',
  },

  /**
   * An enum representing different line types that are available for line annotations
   * @property {string} SOLID - solid line style
   * @property {string} DASH_1_2_2 - dashed line with [1,2,2] style
   * @property {string} DASH_3_3 - dashed line with [3,3] style
   * @property {string} DASH_3_3_4 - dashed line with [3,3,4] style
   * @property {string} DASH_1_3_5 - dashed line with [1,3,5] style
   * @property {string} DASH_2_2_4 - dashed line with [2,2,4] style
   * @property {string} DASH_4_3_16_3 - dashed line with [4,3,16,3] style
   * @property {string} CLOUDY - cloudy line style
   */
  LineStyleType: {
    SOLID: 'SOLID',
    DASH_1_2_2: 'DASH_1_2_2',
    DASH_3_3: 'DASH_3_3',
    DASH_3_3_4: 'DASH_3_3_4',
    DASH_1_3_5: 'DASH_1_3_5',
    DASH_2_2_4: 'DASH_2_2_4',
    DASH_4_3_16_3: 'DASH_4_3_16_3',
    CLOUDY: 'CLOUDY',
  },

  /**
   * Restores the draw function back to the default.
   * @param {Core.Annotations.Annotation} annotationClass - The class (constructor) of the annotation
   */
  restoreDraw(annotationClass) {},

  /**
   * Change ControlHandle's draw to customize appearance on the provided canvas context.
   * @param {Core.Annotations.ControlHandle} controlHandle - The class (constructor) of the controlHandle
   * @param {CustomControlHandleDrawHandler} controlHandleDrawHandler - The handler will customize the appearance of the controlHandle
   * @example
   * Annotations.setCustomControlHandleDrawHandler(Core.Annotations.ControlHandle, function(ctx, annotation, selectionBox, zoom, {controlHandle, originalDraw}) {
    if(controlHandle instanceof Core.Annotations.BoxControlHandle) {
      const dim = this.getDimensions(annotation, selectionBox, zoom);
      ctx.fillStyle = '#FFFFFF';
      ctx.beginPath();
      ctx.moveTo(dim.x1 + (dim.getWidth() / 2), dim.y1);
      ctx.lineTo(dim.x1 + dim.getWidth(), dim.y1 + dim.getHeight());
      ctx.lineTo(dim.x1, dim.y1 + dim.getHeight());
      ctx.closePath();
      ctx.stroke();
      ctx.fill();
    } else {
      originalDraw(ctx, annotation, selectionBox, zoom);
    }
  })
   */
  setCustomControlHandleDrawHandler(controlHandle, controlHandleDrawHandler) {}
}