// @ts-ignore
importScripts("./ComPDFkit.js")

import MessageHandler from "./messageHandler"
import PriorityQueue from "./PriorityQueue"
import TaskQueue from "./TaskQueue"

function sleep(delay: number) {
  return new Promise((resolve) => setTimeout(resolve, delay))
}

declare var Module: {
  _InitDocument: () => number;
  _GetPageCount: (ptr: number) => number;
  _GetComPDFkitLicenseErrorcode: () => number;
  _InitComPDFkit: (licensePtr: number, flag: boolean, devicePtr: number, appIdPtr: number, someInt: number) => void;
  _LoadDocumentByStream: (ptr: number, docId: number, length: number, passwordPtr: number) => number;
  _GetPageSizeForIndex: (ptr: number, pageIndex: number) => void;
  _GetPageWithoutParse: (ptr: number, pageIndex: number) => number;
  _GetPageRotation: (ptr: number) => number;
  _ClearDocument: (ptr: number) => void;
  _ClearPage: (ptr: number) => void;
}

declare var stringToNewUTF8: (str: string) => number

type Page = {
  ptr: number
  pageIndex: number
}

type PageSizes = {
  [key: number]: {
    width: number,
    height: number,
    rotaion: number
  }
}

let ComPDFKitJS: {
  opened_files: Uint8Array[]
} = {
  opened_files: []
}
let PageSize = {
  Width: 0,
  Height: 0
}

class ComPDFKitWorker {
  messageHandler: MessageHandler | null = null
  private _queue: PriorityQueue | null = null
  actionMap: { [key: string]: (...args: any[]) => any } = {}
  actionMapAsync: Map<string, (...args: any[]) => any> = new Map()
  currentAction: string | null = null
  license: string | null = null
  appId: string = ''
  documentViewer: {
    [key: string | number]: {
      ptr: number
      buffer: Uint8Array
      password: string
      pages?: Page[]
    }
  } = {}

  constructor(self: typeof globalThis) {
    const messageHandler = new MessageHandler('worker', 'main', self)
    this.messageHandler = messageHandler

    this.setup()
  }

  setup() {
    this.deferredInit()

    if (!this.messageHandler) return
    const messageHandler = this.messageHandler
    messageHandler.on('test', this.testMessageHandler.bind(this))

    const taskQueue = new TaskQueue()
    messageHandler.onNextAsync(() => {
      taskQueue.addTask(() => {
        if (this.size() && !this.currentAction) {
          const item = this.dequeue()
          this.handleAsync(item)
        }
      })
    })

    messageHandler.on('InitConfig', this.initConfig.bind(this))
    this.onAsync('NewDocument', this.newDocument.bind(this))
    messageHandler.on('PasswordRequest', this.requestPassword.bind(this))
    messageHandler.on('CloseDocument', this.closeDocument.bind(this))
  }

  deferredInit() {
    this._queue = new PriorityQueue({
      comparator: (e, w) => {
        return e.priority === w.priority ? e.callbackId - w.callbackId : w.priority - e.priority
      }
    })
  }

  verifyLicense() {
    if (this.license) {
      const appId = this.appId || ''
      let licensePtr = stringToNewUTF8(this.license)
      const devicePtr = stringToNewUTF8('')
      let appIdPtr = stringToNewUTF8(appId)
      Module._InitComPDFkit(licensePtr, false, devicePtr, appIdPtr, 0)
      const result = licenseErrorMessage()
      return result
    }
  }

  initConfig(data: {
    license: string
    appId: string
  }) {
    const { license, appId } = data

    this.license = license
    this.appId = appId
    const result = this.verifyLicense()
    return result
  }

  queue(value: any) {
    this._queue?.queue(value)
  }

  dequeue() {
    this._queue?.dequeue()
  }

  peek() {
    this._queue?.peek()
  }

  remove(value: any) {
    return this._queue?.remove(value)
  }

  removeAllMatching(test: Function, callback: Function) {
    return this._queue?.removeAllMatching(test, callback)
  }

  size() {
    return this._queue?.size()
  }

  asyncCallCapability(data: any) {
    data.asyncCallCapability = Promise.withResolvers()
    this.currentAction || this.size() ? this.queue(data) : this.handleAsync(data)
    return data.asyncCallCapability.promise
  }

  handleAsync(event: any) {
    const { data, action } = event
    this.currentAction = event
    const capability = data.asyncCallCapability;

    this.actionMap[action](data, event).then((data: any) => {
      this.currentAction = null
      capability.resolve(data)
    }, (error: any) => {
      this.currentAction = null
      capability.reject(error)
    })
  }

  onAsync(actionName: string, handler: (...args: any[]) => any) {
    this.messageHandler?.onAsync(actionName, this.asyncCallCapability)
    this.actionMap[actionName] = handler
  }

  testMessageHandler(data: any) {
    const messageHandler = this.messageHandler
    if (data.array instanceof Uint8Array) {
      const supportTransfers = 255 === data.array[0];

      messageHandler?.send('test', {
        supportTransfers
      })
    } else {
      messageHandler?.send("test", !1)
    }
  }

  newDocument(data: {
    docId: number
    password: string
    buffer: Uint8Array
  }) {
    const { password, docId, buffer } = data
    const ptr = Module._InitDocument()

    this.documentViewer[docId] = {
      ptr,
      buffer,
      password,
    }

    const resultCode = this.loadDocument({
      ptr,
      buffer,
      password
    })

    if (resultCode === 4) {
      throw {
        docId,
        message: password ? 'Wrong password' : 'Password required',
        type: 'Password Required'
      }
    } else if (resultCode === 0) {
      const pageSizes = this.getPageSizes(ptr)
      return {
        docId,
        pageSizes
      }
    } else {
      throw {
        docId,
        message: 'Unknown error',
        type: 'Unknown Error'
      }
    }
  }

  getPageSizes(ptr: number): PageSizes {
    const pageCount = Module._GetPageCount(ptr)
    const pageSizes: {
      [key: number]: {
        width: number,
        height: number,
        rotaion: number
      }
    } = {}
    for (let pageIndex = 0; pageIndex < pageCount; pageIndex++) {
      PageSize.Width = 0
      PageSize.Height = 0
      Module._GetPageSizeForIndex(ptr, pageIndex)
      const pagePtr = Module._GetPageWithoutParse(ptr, pageIndex)
      const rotaion = Module._GetPageRotation(pagePtr)
      pageSizes[pageIndex] = {
        width: PageSize.Width,
        height: PageSize.Height,
        rotaion
      }
    }

    return pageSizes
  }

  requestPassword(data: {
    docId: number
    password: string
  }) {
    const { docId, password } = data
    const document = this.documentViewer[docId]
    document.password = password

    const { ptr, buffer } = document
    const resultCode = this.loadDocument({
      ptr,
      buffer,
      password
    })

    if (resultCode === 4) {
      return {
        docId,
        pass: false
      }
    } else if (resultCode === 0) {
      const pageSizes = this.getPageSizes(ptr)
      return {
        docId,
        pass: true,
        pageSizes
      }
    } else {
      throw {
        docId,
        message: 'Unknown error',
        type: 'Unknown Error'
      }
    }
  }

  loadDocument(data: {
    ptr: number
    buffer: Uint8Array
    password: string
  }) {
    const { ptr, buffer, password } = data
    ComPDFKitJS.opened_files[ComPDFKitJS.opened_files.length] = buffer

    const length = buffer.length
    const passwordPtr = stringToNewUTF8(password)
    const resultCode = Module._LoadDocumentByStream(ptr, 0, length, passwordPtr)
    ComPDFKitJS.opened_files.pop()

    return resultCode
  }

  closeDocument(data: {
    docId: number
  }) {
    const docId = data.docId
    const item = this.documentViewer[docId]
    Module._ClearDocument(item.ptr)
    delete this.documentViewer[docId]
    this.removeAllMatching((item: any) => item.data && item.data.docId === docId, (item: any) => {
      item.asyncCallCapability.reject('Document closed')
    })
  }

  clearPages(docId: number) {
    const document = this.documentViewer[docId]
    if (document && document.pages && document.pages.length) {
      document.pages.forEach((page: Page) => {
        this.clearPage(page.ptr)
      })
    }
  }

  clearPage(ptr: number) {
    Module._ClearPage(ptr)
  }
}

function setUp(): Promise<void> {
  // @ts-ignore
  while (!runtimeInitialized) {
    sleep(100)
  }
  return Promise.resolve()
}

setUp().then(() => {
  const ready = (global: typeof globalThis) => {
    global.onmessage = (e: MessageEvent) => {
      const action = e.data.action
      if (action === 'init') {
        new ComPDFKitWorker(global)
      }
    }
  }

  ready(typeof window !== 'undefined' ? window : self)
})

function licenseErrorMessage(feature?: string) {
  const code = Module._GetComPDFkitLicenseErrorcode()
  let result: {
    code: number
    message?: string
  } = {
    code,
  }
  if (code) {
    switch (code) {
      case -100:
        result.message = 'The license is invalid'
        break
      case -101:
        result.message = 'The license is expired'
        break
      case -102:
        result.message = 'The license does not support the current platform'
        break
      case -103:
      case -104:
        result.message = 'The license does not support your domain'
        break
      case -105:
        result.message = 'The license does not support the ' + feature
        break
      case -106:
        result.message = 'License has not been initialized'
        break
      case -1000:
        result.message = 'Illegal access to the API interface'
        break
      case -1001:
        result.message = 'Failed to read license file'
        break
    }
  }
  return result
}
