import Color from "./Color";
import { v4 as uuidv4 } from 'uuid';

/**
 * Represents an annotation.
 * @memberof Core.Annotations
 */
class Annotation {

  /**
   * The base class for all annotations
   * @param {object} [initializer] - A map of values to auto-initialize certain properties of the annotation. You can only initialize properties defined on the annotation under the Members section (unless specified otherwise).
   */
  constructor(initializer) {
    /** The author of the annotation. */
    this.Author = initializer?.Author || 'Guest';

    /** Gets or sets the annotation's stroke color. */
    this.Color = initializer?.Color || new Color(0, 0, 0, 1);

    /** Gets or sets the date that the annotation was last modified. */
    this.DateModified = initializer?.DateModified || new Date();

    /** Gets or sets the name of the annotation element in XFDF.
     * 
     * This property cannot be initialized through the initializer.
     */
    this.elementName = 'annotation';

    /** Gets or sets whether the annotation is hidden. */
    this.Hidden = initializer?.Hidden || false;

    /** A unique identifier for the annotation. Corresponds to the name attribute in XFDF. */
    this.Id = initializer?.Id || uuidv4();

    /** The Id of the annotation that this annotation is replying to or the parent annotation it is grouped with. */
    this.InReplyTo = initializer?.InReplyTo || null;

    /** Gets or sets whether the annotation is invisible, only if it is an unknown annotation type. Generally for hiding annotations you should use "Hidden". */
    this.Invisible = initializer?.Invisible || false;

    /** Gets or sets whether the annotation has been added since the last time the AnnotCommand was retrieved. */
    this.IsAdded = initializer?.IsAdded || false;

    /** Gets or sets whether any parts of the annotation drawn outside of the rect are clickable. */
    this.IsClickableOutsideRect = initializer?.IsClickableOutsideRect || false;

    /** Gets or sets whether the annotation has been deleted since the last time the AnnotCommand was retrieved. */
    this.IsDeleted = initializer?.IsDeleted || false;

    /** Gets or sets whether the annotation has been modified since the last time the AnnotCommand was retrieved. */
    this.IsModified = initializer?.IsModified || false;

    /** Gets or sets the page number of a document that the annotation appears on. */
    this.PageNumber = initializer?.PageNumber || 1;

    /** The type of reply. */
    this.ReplyType = initializer?.ReplyType;

    /** Gets or sets a selection model constructor for how an annotation gets selected. */
    this.selectionModel = initializer?.selectionModel;

    /** Represents the subject of the annotation. Default is the annotation's type. */
    this.Subject = initializer?.Subject;

    /** Gets or sets whether the ToggleNoView flag is set on the annotation. */
    this.ToggleNoView = initializer?.ToggleNoView;

    /** The name of the tool that is used to create this annotation. */
    this.ToolName = initializer?.ToolName;

    /** Gets or sets the width of the annotation. */
    this.Width = initializer?.Width || 0;

    /** Gets or sets the height of the annotation. */
    this.Height = initializer?.Height || 0;

    /** Gets or sets the annotation's x-axis position. */
    this.X = initializer?.X || 0;

    /** Gets or sets the annotation's y-axis position. */
    this.Y = initializer?.Y || 0;
  }

  /**
   * Gets the text content for the annotation. Contents may be displayed in an annotation's popup or directly on the page (in the case of FreeTextAnnotation).
   * @returns {string} the text content for the annotation.
   */
  getContents() {
    return this.content;
  }

  /**
   * Returns custom data associated with the given key.
   * @param {string} key - The key for which to retrieve the associated data.
   * @returns {string} value The custom data. If no data is available an empty string is returned.
   */
  getCustomData(key) {
    return this[key] || '';
  }

  /**
   * Gets the height of the annotation.
   * @returns {number} the height of the annotation.
   */
  getHeight() {
    return this.Height;
  }

  /**
   * Gets the page number of the annotation. Note: page number starts from 1.
   * @returns {number} The annotation's page number.
   */
  getPageNumber() {
    return this.PageNumber;
  }

  /**
   * Get annotation bounding rectangle
   * @returns {Core.Math.Rect} The annotation's bounding rectangle.
   */
  getRect() {
    return this.rect;
  }

  /**
   * Gets the list of replies to this annotation.
   * @returns {Array.<Core.Annotations.Annotation>} The list of replies
   */
  getReplies() {
    return this.replies;
  }

  /**
   * Gets the status of the annotation, and returns an empty string if no status set.
   * @returns {string} The last status update.
   */
  getStatus() {
    return this.status || '';
  }

  /**
   * Gets the width of the annotation.
   * @returns {number} The width of the annotation.
   */
  getWidth() {
    return this.Width;
  }

  /**
   * Gets the x position measured in page coordinates of an annotation.
   * @returns {number} the x position
   */
  getX() {
    return this.X;
  }

  /**
   * Gets the y position measured in page coordinates of an annotation.
   * @returns {number} the y position
   */
  getY() {
    return this.Y;
  }

  /**
   * Returns true if annotation is part of the original document, false otherwise.
   * @return {boolean} Whether the annotation is internal to the document or not
   */
  isInternal() {
    return this.isInternal || false;
  }

  /**
   * Gets whether the annotation is a reply to another annotation.
   * @returns {boolean} Returns true if it does reply to another annotation, false otherwise.
   */
  isReply() {
    return !!this.ReplyType;
  }

  /**
   * Whether the annotation is visible on the document or not **when viewing**. If the Hidden or NoView flags are set, or if the annotation is a reply to another annotation then it won't be visible.
   * @returns {boolean} Whether the annotation is visible on the document or not when viewing
   */
  isVisible() {
    return !this.Invisible;
  }

  /**
   * Set the text content for the annotation. Note that this will not refresh the text in the UI.
   * @param {string} value - the text content to be set
   */
  setContents(value) {
    this.content = value;
  }

  /**
   * Sets the custom data associated with the specified key.
   * @param {string} key - The key under which to store this custom data. Data will automatically be saved in the exported XFDF and merged into the PDF when downloading.
   * @param {string} value - The custom data to store
   */
  setCustomData(key, value) {
    this[key] = value;
  }

  /**
   * Sets the height of the annotation.
   * @param {number} value - the height of the annotation.
   */
  setHeight(value) {
    this.height = value;
  }

  /**
   * Sets the page number of the annotation. Note: page numbers start from 1.
   * @param {number} value - the page number to be set
   */
  setPageNumber(value) {
    this.pageNumber = value;
  }

  /**
   * Sets the size and location of the annotation's bounding rectangle. Use this method instead of resize when only the x, y, width and height needs to be modified. Beware: this method ignores coordinates if they are the wrong way around. (It hasn't been fixed for reasons of maintaining backward compatibility. use setRectWithNormalization instead if you want it to reverse the wrong-way-round coordinates instead of ignoring them.)
   * @param {Core.Math.Rect} rect - the new bounding rectangle
   */
  setRect(rect) {
    this.rect = rect;
  }

  /**
   * Sets the width of the annotation.
   * @param {number} value - the width of the annotation.
   */
  setWidth(value) {
    this.Width = value;
  }

  /**
   * Sets the x position measured in page coordinates of an annotation.
   * @param {number} value - the x position
   */
  setX(value) {
    this.X = value;
  }

  /**
   * Sets the y position measured in page coordinates of an annotation.
   * @param {number} value - the y position
   */
  setY(value) {
    this.Y = value;
  }
}

export default Annotation;