/**
 * Represents a class that contains a color's RGB and alpha value.
 * @memberof Core.Annotations
 */
class Color {

  /**
   * Creates a new instance of Color.
   * @param {number} r - the R (red) value (0-255)
   * @param {number} g - the G (green) value (0-255)
   * @param {number} b - the B (blue) value (0-255)
   * @param {number} [a] - the A (alpha) value (0-1.0)
   * 
   * @property {number} R - the R (red) value (0-255)
   * @property {number} G - the G (green) value (0-255)
   * @property {number} B - the B (blue) value (0-255)
   * @property {number} A - the A (alpha) value (0-1.0)
   */
  constructor(r = 0, g = 0, b = 0, a) {
    this.R = Number(r);
    this.G = Number(g);
    this.B = Number(b);
    this.A = (a == 0 || a) ? Number(a) : 1;
  }

  /**
   * Returns the color as a hex string e.g. #FFFFFF
   * @returns {string} The hex color string.
   */
  toHexString() {
    const r = this.R;
    const g = this.G;
    const b = this.B;
    const hexString = ((1 << 24) | (r << 16) | (g << 8) | b).toString(16).slice(1).toUpperCase();
    return `#${hexString}`;
  }

  hexToRgb(hexColor) {
    const hex = hexColor.slice(1);
    const bigint = parseInt(hex, 16);
    const R = (bigint >> 16) & 255;
    const G = (bigint >> 8) & 255;
    const B = bigint & 255;
    return {
      R,
      G,
      B
    }
  }

  /**
   * Outputs the current color as a CSS3 RGB color string.
   * @returns {string} The CSS3 RGB color string.
   * @example
   * ex. "rgb(0,0,0)" or "rgba(0,255,0,0.5)"
   */
  toString() {
    if (this.A === 1) return `rgb(${this.R}, ${this.G}, ${this.B})`;
    return `rgba(${this.R}, ${this.G}, ${this.B}, ${this.A})`;
  }

  setColor(r, g, b, a) {
    this.R = r;
    this.G = g;
    this.B = b;
    typeof a === 'number' && (this.A = a);
  }

  formatColor(value) {
    const format = this.getColorFormat(value);
    if (format === 'hex') {
      const { R, G, B } = this.hexToRgb(value);
      return {
        R,
        G,
        B
      };
    } else if (format === 'rgb') {
      const match = value.match(/(\d+),\s*(\d+),\s*(\d+)/);
      const R = parseInt(match[1]);
      const G = parseInt(match[2]);
      const B = parseInt(match[3]);
      return {
        R,
        G,
        B
      };
    } else {
      console.error({
        message: 'Invalid color value'
      });
    }
  }

  getColorFormat(colorString) {
    if (colorString.startsWith('#')) {
      return 'hex';
    } else if (colorString.startsWith('rgb(')) {
      return 'rgb';
    } else {
      return 'unknown';
    }
  }

  convertColorToCppFormat() {
    const R = roundToDecimalPlaces(this.R / 255);
    const G = roundToDecimalPlaces(this.G / 255);
    const B = roundToDecimalPlaces(this.B / 255);
    return {
      R,
      G,
      B
    };
  }
}

export default Color;
