import Dispatcher from '../../Actions/Dispatcher';

/**
 * @memberof Core.Annotations.Forms
 * @mixes Core.Actions.Dispatcher
 * @listens Core.Annotations.Forms.Field#event:calculate
 * @listens Core.Annotations.Forms.Field#event:commit
 */
class Field {

  /**
   * Represents a PDF Form field.
   * @param {string} name - The field's full name.
   * @param {Core.Annotations.Forms.Field|object} [options] - The options with which to construct the field. If options is a Field, the constructor returns the existing object.
   * @param {string} [options.type] - The field's type. One of ('Tx', 'Btn', 'Ch' or 'Sig').
   * @param {string} [options.value] - The value of the field
   * @param {Core.Annotations.WidgetFlags} [options.flags] - The flags for the field
   * @param {Array.<Core.Annotations.Forms.Field>} [options.children] - The field's children.
   * @param {Array.<Core.Annotations.WidgetAnnotation>} [options.widgets] - The field's child widgets.
   * @param {Core.Annotations.Forms.FieldManager} [options.fieldManager] - The field's field manager.
   * @param {Core.Annotations.Font} [options.font] - The font that is used by the field.
   * @param {number} [options.maxLen] - The field's maximum length. -1 means infinite.
   * @param {string} [options.tooltipName] - The text to be displayed when hovering over the field.
   * @param {Array.<object>} [options.options] - The field's options (or permanently null if the field type does not support options).
   * 
   * @property {string} type - Returns the field's type. One of ('Tx', 'Btn', 'Ch' or 'Sig')
   * @property {Array.<Core.Annotations.Forms.Field>} [children=[]] - The field's children
   * @property {Array.<Core.Annotations.WidgetAnnotation>} [widgets=[]] - The field's child widgets
   * @property {object.<string, Array.<Core.Actions.Action>>} [actions={}] - Gets the field's actions.
   * @property {number|string} [defaultValue=null] - The field's default value
   * @property {number|string} [exportValue=null] - The field's export value. Defaults to the field's value if not set.
   * @property {Core.Annotations.WidgetFlags} [flags=new Core.Annotations.WidgetFlags()] - The field's flags object
   * @property {Core.Annotations.Font} [font=new Core.Annotations.Font()] - Returns the field's associated font object
   * @property {boolean} [IsModified=false] - Is field is modified?
   * @property {number} [maxLen=-1] - The field's maximum length. -1 means infinite.
   * @property {string} name - The full name of the field.
   * @property {Array.<object>} [options=[]] - The field's options (or permanently null if the field type does not support options).
   * @property {string} [quadding='Left-justified'] - Returns the field's quadding. (One of 'Left-justified', 'Right-justified', 'Centered')
   * @property {number|string} [value=null] - The field's value.
   * @property {string} tooltipName - The text to be displayed when hovering over the field.
   */
  constructor(name, options) {
    this.name = name;
    this.type = options?.type || '';
    this.children = options?.children || [];
    this.widgets = options?.widgets || [];
    this.actions = options?.actions || {};
    this.defaultValue = options?.defaultValue || null;
    this.exportValue = options?.exportValue || null;
    this.flags = options?.flags || new Core.Annotations.WidgetFlags();
    this.font = options?.font || new Core.Annotations.Font();
    this.IsModified = options?.IsModified || false;
    this.maxLen = options?.maxLen || -1;
    this.quadding = options?.quadding || 'Left-justified';
    this.options = options?.options || [];
    this.value = options?.value || null;
    this.tooltipName = options?.tooltipName;

    Object.assign(this, Dispatcher)
  }

  /**
   * Commits the new value to the field (triggers Validate action, and updates all widgets)
   * @param {*} event - Ignored.
   * @param {string|number} [value] - The new value.
   * @param {Core.Annotations.WidgetAnnotation} [widget] - Optionally a particular widget annotation associated with the field
   */
  commit(event, value, widget) {}

  /**
   * Returns the field's type
   * @return {string} - The field type as defined by the FormFieldTypes enum
   */
  getFieldType() {}

  /**
   * Gets the field's value
   * @return {string|number} - The value of the field
   */
  getValue() {}

  /** Disable visibility of all child widgets */
  hide() {}

  /**
   * Checks whether the current field is a terminal/leaf node (no children).
   * @return {boolean} - Whether this field is terminal/leaf node
   */
  isTerminal() {}

  /** Refresh the appearance of each widget in the current field. */
  refreshAppearances() {}

  /**
   * Set all field properties form the given object.
   * @param {Core.Annotations.Forms.Field|object} options - The object to set properties from.
   */
  set(options) {
    const self = this;
    Object.keys(options).forEach(key => {
      if (key in self) self[key] = options[key];
    });
  }

  /**
   * Sets the field's type. Tries to retain as much type-specific information as possible across type changes.
   * @param {string} type - The new field type, one of Tx, Ch, Btn or Sig.
   */
  setType(type) {
    this.type = type;
  }

  /**
   * Sets the field's value, following any type-specific constraints, and updating all widgets.
   * @param {?(string|number)} value - The value to set.
   * @param {Core.Annotations.WidgetAnnotation} [widget] - Optionally a particular widget annotation associated with the field
   * @fires Core.Annotations.Forms.Field#event:change
   */
  setValue(value, widget) {}

  /** Enable visibility of all child widgets */
  show() {}

  /**
   * A signal that the field should fire its own calculation events.
   * @event Core.Annotations.Forms.Field#calculate
   * @param {Core.Annotations.Forms.Field} source - The field that caused the calculation event.
   */
  calculate(source) {}

  /**
   * An event that represents a field's value having changed.
   * @event Core.Annotations.Forms.Field#change
   * @param {Core.Annotations.Forms.Field} field - The field that has changed.
   */
  change() {}

  /**
   * The field's value has been changed by the user. Sets a new value.
   * @event Core.Annotations.Forms.Field#commit
   * @param {?(string|number)} value - The new field value.
   */
  commit(value) {}
}

export default Field;
