import IPathAnnotation from './IPathAnnotation';

/** 
 * @memberof Core.Annotations
 */
class FreeTextAnnotation extends IPathAnnotation {

  /**
   * An enum with possible freetext intents.
   * @property {string} FreeText - Treat as regular freetext
   * @property {string} FreeTextCallout - Treat as callout
   */
  static Intent = {
    FreeText: 'FreeText',
    FreeTextCallout: 'FreeTextCallout'
  };

  /**
   * Represents a free text annotation.
   * @extends Core.Annotations.IPathAnnotation
   * 
   * @param {string} [intent] - The intended usage type of this freetext annotation. Could either be freetext or callout.
   * @param {object} [initializer] - A map of values to auto-initialize certain properties of the annotation. You can only initialize properties defined on the annotation under the Members section (unless specified otherwise).
   * 
   * @property {boolean} EXPORT_CALCULATED_FONT_SIZE - Whether to export calculated font size for auto-sized fonts. Default: true
   * @property {object} Intent - An enumeration of possible intents for FreeText Annotations
   * @property {string} Intent.FreeText - Should treat as regular freetext
   * @property {string} Intent.FreeTextCallout - Should treat as callout
   */
  constructor(intent, initializer = {}) {
    super(initializer);
    this.EXPORT_CALCULATED_FONT_SIZE = true;
    this.Intent = intent || FreeTextAnnotation.Intent.FreeText;

    Object.keys(initializer).forEach(key => {
      if (this.hasOwnProperty(key)) {
        this[key] = initializer[key];
      }
    });

    /** The font to use for the annotation's text. */
    this.Font = '';

    /** The font size to use for the annotation's text, specified like a CSS font size. */
    this.FontSize = 14;

    /** The horizontal alignment of the annotation's text (left, right, center) */
    this.TextAlign = 'left';

    /** The color of the text in the annotation. */
    this.TextColor = '';

    /** The vertical alignment of the annotation's text (top, bottom, center) default: top */
    this.TextVerticalAlign = 'top';

    this.padding = 8;
  }

  /**
   * Gets the intent of the FreeText annotation e.g. FreeTextCallout
   * @return {string} The intent of the annotation
   */
  getIntent() {
    return this.Intent;
  }

  /**
   * Get the rectangle difference of the annotation bounding rect and the text rect
   * @return {Core.Math.Rect} the rectangle difference
   */
  getPadding() {
    return this.padding;
  }

  /**
   * Sets the intent of the FreeText annotation
   * @param {string} intent - The new intent for the annotation
   */
  setIntent(intent) {
    this.Intent = intent;
  }

  /**
   * Sets the rectangle difference of the annotation bounding rect and the text rect
   * @param {Core.Math.Rect} rect - The new rectangle difference
   */
  setPadding(rect) {}
}

export default FreeTextAnnotation;
