import MarkupAnnotation from './MarkupAnnotation';

/**
 * @memberof Core.Annotations
 */
class LineAnnotation extends MarkupAnnotation {

  /**
   * @extends Core.Annotations.MarkupAnnotation
   * @property {string} START - Indicates the caption is snapped to start point of measurement annotation.
   * @property {string} CENTER - Indicates the caption is snapped to center point of measurement annotation.
   * @property {string} END - Indicates the caption is snapped to end point of measurement annotation.
   */
  static CaptionSnapPositions = {
    START: 'start',
    CENTER: 'center',
    END: 'end'
  }
  
  /**
   * Represents a line annotation.
   * @extends Core.Annotations.MarkupAnnotation
   * @param {object} [initializer] - A map of values to auto-initialize certain properties of the annotation. You can only initialize properties defined on the annotation under the Members section (unless specified otherwise).
   */
  constructor(initializer) {
    super(initializer);

    /** Gets or sets the start point of the line. */
    this.Start = initializer?.Start;

    /** Gets or sets the end point of the line. */
    this.End = initializer?.End;

    /** Gets or sets the annotation's scale from its measure dictionary. */
    this.Scale = initializer?.Scale || 1;

    /** Gets or sets the border style of an annotation. e.g Solid, Cloudy */
    this.Style = initializer?.Style || 'solid';
  }

  /**
   * Gets the ending point of the line
   * @return {Core.Math.Point} The end point
   */
  getEndPoint() {
    return this.End;
  }

  /**
   * Returns the length of the line in points
   * @return {number} The length in points
   */
  getLineLength() {
    return Math.sqrt(Math.pow(this.End.x - this.Start.x, 2) + Math.pow(this.End.y - this.Start.y, 2));
  }

  /**
   * Gets the starting point of the line
   * @rerturn {Core.Math.Point} The start point
   */
  getStartPoint() {
    return this.Start;
  }

  /**
   * Sets the ending point of the line
   * @param {number} x - The x coordinate of the point
   * @param {number} y - The y coordinate of the point
   */
  setEndPoint(x, y) {
    this.End.x = x;
    this.End.y = y;
  }

  /**
   * Sets the starting point of the line
   * @param {number} x - The x coordinate of the point
   * @param {number} y - The y coordinate of the point
   */
  setStartPoint(x, y) {
    this.Start.x = x;
    this.Start.y = y;
  }
}

/**
 * the point in viewer coordinates
 * @typedef {Object} Core.Annotations.LineAnnotation.LeaderPoints
 * @property {Core.Math.Point} Start - the start point
 * @property {Core.Math.Point} End - the end point
 */

export default LineAnnotation;
