/**
 * Represents a class that contains information about how an annotation should behave when selected.
 * @memberof Core.Annotations
 */
class SelectionModel {

  /** Defines the default color for the annotation selection outline when the user is not permitted to make modifications. */
  static defaultNoPermissionSelectionOutlineColor;

  /** Defines the default color for the annotation selection outline. */
  static defaultSelectionOutlineColor;

  /** Defines padding for selection accuracy. Default is 2. Increase this value to make selection more forgiving. */
  static selectionAccuracyPadding = 2;

  /** Defines the dash size for the selection outline. Default is 4; */
  static selectionOutlineDashSize = 4;

  /** Defines the default padding between selection control points. The padding only appears when the control points are too close. */
  static selectionOutlineExtraPadding;

  /** Defines thickness of the annotation selection outline. Default is 1. */
  static selectionOutlineThickness = 1;

  /** Defines if extra padding should be shown when the selected annotation is too small. Default is true. */
  static showPaddingWhenAnnotationIsSmall = true;

  /** Defines if a dashed line style should be used for the selection outline. Default is false. */
  static useDashedLine = false;

  /**
   * Create a new selection model.
   * @param {Core.Annotations.Annotation} annotation - The annotation selected
   * @param {boolean} canModify - Modification of the annotation is allowed
   * @param {boolean} isSelected - The annotation is already selected
   * @param {Core.DocumentViewer} docViewer - An instance of DocumentViewer.
   */
  constructor(annotation, canModify, isSelected, docViewer) {
    this.annotation = annotation;
    this.canModify = canModify;
    this.isSelected = isSelected;
    this.docViewer = docViewer;
  }

  /**
   * Overwrites the functions in SelectionModel(ex: drawSelectionOutline, getDimensions, testSelection).
   * @param {Core.Annotations.SelectionModel} selectionModelClass - The class (constructor) of the selectionModel
   * @param {Core.Annotations.SelectionModel.SelectionModelCustomHandlers} selectionModelCustomHandlers - An object containing multiple handlers to overwrite functions on a SelectionModel.
   * @example
const { Annotations } = instance.Core;
const { SelectionModel, BoxSelectionModel } = Annotations;
SelectionModel.setCustomHandlers(BoxSelectionModel, {
  // draws a diagonal dashed along across the middle of the selected annotation
   drawSelectionOutline(ctx, annotation, zoom, pageMatrix, { selectionModel, originalDrawSelectionOutline }) {
     if (!(annotation instanceof Annotations.RectangleAnnotation)) {
         originalDrawSelectionOutline(ctx, annotation, zoom, pageMatrix);
         return;
       }
      if (typeof zoom !== 'undefined') {
         ctx.lineWidth = SelectionModel.selectionOutlineThickness / zoom;
       } else {
         ctx.lineWidth = SelectionModel.selectionOutlineThickness;
       }
       if (selectionModel.canModify()) {
         ctx.strokeStyle = SelectionModel.defaultSelectionOutlineColor.toString();
       } else {
         ctx.strokeStyle = SelectionModel.defaultNoPermissionSelectionOutlineColor.toString();
       }
       ctx.beginPath();
       ctx.moveTo(annotation.X, annotation.Y);
       ctx.lineTo(annotation.X + annotation.Width, annotation.Y + annotation.Height);
       ctx.closePath();
       ctx.stroke();
       const dashUnit = SelectionModel.selectionOutlineDashSize / zoom;
       const sequence = [dashUnit, dashUnit];
       ctx.setLineDash(sequence);
       ctx.strokeStyle = 'rgb(255, 255, 255)';
       ctx.stroke();
     },
    // Get the dimension that is extended by 8 both horizontally and vertically
    getDimensions(annotation, { selectionModel, originalGetDimensions }) {
     if (!(annotation instanceof Annotations.RectangleAnnotation)) {
        return originalGetDimensions(annotation);
     }
     const x = annotation.X - 4;
     const y = annotation.Y - 4;
     const width = annotation.Width + 2 * 4;
     const height = annotation.Height + 2 * 4;
     return new Annotations.Rect(x, y, x + width, y + height);
    },
    testSelection(annotation, x, y, pageMatrix, zoom, rotation, { selectionModel, originalTestSelection }) {
      if (annotation instanceof Annotations.RectangleAnnotation) {
        return originalTestSelection(annotation, x, y, pageMatrix, zoom, rotation);;
      }
      return Annotations.SelectionAlgorithm.boundingRectTest(annotation, x, y, zoom);
    }
})
   */
  static setCustomHandlers(selectionModelClass, selectionModelCustomHandlers) {}

  /**
   * Defines a function to set selection padding on specific types of annotations.
   * @param {function} fn - A function that returns the padding value for specific types of annotations
   * @example
const { Annotations } = instance.Core;
Annotations.SelectionModel.setSelectionModelPaddingHandler((annotation) => {
  if (annotation instanceof Annotations.FreeTextAnnotation) {
    return 30;
  }
  if (annotation instanceof Annotations.RectangleAnnotation) {
    return 20;
  }
  return 0;
});
   */
  static setSelectionModelPaddingHandler(fn) {}

  /**
   * Indicates that the associated annotation is able to be modified
   * @return {boolean}
   */
  canModify() {
    return this.canModify;
  }

  /**
   * Draws the selection outline of the annotation. By default, a rectangle is drawn based on the annotations x, y, width and height.
   * @param {CanvasRenderingContext2D} ctx 
   * @param {Core.Annotations.Annotation} annotation 
   * @param {number} zoom 
   */
  drawSelectionOutline(ctx, annotation, zoom) {}

  /**
   * Returns the ControlHandle objects associated with this selection model.
   * @return {Array.<Core.Annotations.ControlHandle>} An array of ControlHandleObject
   */
  getControlHandles() {}

  /**
   * Gets the dimensions {x, y, width, height} of the selection bounding box. It may be different from the annotation's bounding box. e.g. The selection bounding box may have a padding.
   * @param {Core.Annotations.Annotation} annotation
   * @return {Core.Math.Rect}
   */
  getDimensions(annotation) {}

  /**
   * Indicates that the associated annotation is already selected. This can be useful to implement different selection behaviors when an annotation is selected.
   * @returns {boolean}
   */
  isSelected() {
    return this.isSelected;
  }

  /**
   * Hit detection for each control handle.
   * @param {Core.Annotations.Annotation} annotation 
   * @param {number} zoom 
   * @param {number} x 
   * @param {number} y 
   * @return {Core.Annotations.ControlHandle} The control handle that was hit
   */
  testControlHandles(annotation, zoom, x, y) {}

  /**
   * Determines if the provided point is a hit on the selected annotationhandle. * See {@link Core.Annotations.SelectionAlgorithm} for usuable selection algorithms.
   * @param {Core.Annotations.Annotation} annotation - the annotation
   * @param {number} x - the x-coordinate of the point to test, in page coordinates
   * @param {number} y - the y-coordinate of the point to test, in page coordinates
   * @param {object} pageMatrix - the page matrix of the page the annotation is on
   * @param {number} zoom - the zoom level of the page the annotation is on
   * @param {Core.PageRotation} rotation - the rotation of the page the annotation is on
   * @return {boolean} true if the provided point is a hit
   */
  testSelection(annotation, x, y, pageMatrix, zoom, rotation) {}
}

/**
 * Callback that gets passed to drawSelectionOutlineHandler in setCustomHandlers.
 * @typedef {function} Core.Annotations.SelectionModel.CustomDrawSelectionOutlineHandler
 * @param {CanvasRenderingContext2D} ctx - A canvas context
 * @param {Core.Annotations.Annotation} annotation - The annotation being selected
 * @param {number} zoom - the current zoom level of the document
 * @param {object} pageMatrix - The transformation matrix for the page that the annotation is on
 * @param {object} options - Additional options and parameters
 * @param {Core.Annotations.SelectionModel} options.selectionModel - The selection model of the annotation which is currently selected.
 * @param {function} options.originalDrawSelectionOutline - The original draw function of this selection model
 */
function CustomDrawSelectionOutlineHandler(ctx, annotation, zoom, pageMatrix, options) {}

/**
 * Callback that gets passed to getDimensionsHandler in setCustomHandlers.
 * @typedef {function} Core.Annotations.SelectionModel.CustomGetDimensionsHandler
 * @param {Core.Annotations.Annotation} annotation - The annotation being selected
 * @param {Core.Math.Rect} selectionBox
 * @param {number} zoom - the current zoom level of the document
 * @param {object} options - Optional options
 * @param {Core.Annotations.SelectionModel} options.selectionModel - The selection model of the annotation which is currently selected.
 * @param {function} options.originalGetDimensions - The original getDimensions function of this selection model
 * @return {Core.Math.Rect}
 */
function CustomGetDimensionsHandler(annotation, selectionBox, zoom, options) {}

/**
 * Callback that gets passed to testSelectionHandler in setCustomHandlers.
 * @typedef {function} Core.Annotations.SelectionModel.CustomTestSelectionHandler
 * @param {number} x - The x-coordinate of the point to test, in page coordinates
 * @param {number} y - The y-coordinate of the point to test, in page coordinates
 * @param {object} pageMatrix - the page matrix of the page the annotation is on
 * @param {number} zoom - The zoom level of the page the annotation is on
 * @param {Core.PageRotation} rotation - The rotation of the page the annotation is on
 * @param {object} options - Optional options
 * @param {Core.Annotations.SelectionModel} options.selectionModel - The selection model of the annotation which is currently selected.
 * @param {function} options.originalTestSelection - The original draw function of this selection model
 * @return {boolean}
 */
function CustomTestSelectionHandler(x, y, pageMatrix, zoom, rotation, options) {}

/**
 * Callback that gets passed to testSelectionHandler in setCustomHandlers.
 * @typedef {object} Core.Annotations.SelectionModel.SelectionModelCustomHandlers
 * @property {Core.Annotations.SelectionModel.CustomDrawSelectionOutlineHandler} [drawSelectionOutline] - Changes how a selection model is drawn within WebViewer.
 * @property {Core.Annotations.SelectionModel.CustomGetDimensionsHandler} [getDimensions] - Gets the rect of the selection bounding box.
 * @property {Core.Annotations.SelectionModel.CustomTestSelectionHandler} [testSelection] - Determines if the provided point is a hit on the selected annotation handle.
 */

export default SelectionModel;
