/**
 * @memberof Core
 */
class EventHandler {
  constructor() {
    this._listeners = Object.create(null);
  }

  /**
   * Add a handler to the given event name
   * @param {string|number} type - The name of the event to listen to
   * @param {function} fn - The handler to be called when the event is triggered
   * @param {object} [options] - Optional options object for addEventListener
   * @param {boolean} options.once - If true then the handler will be called only once
   * @returns {object} Returns the object that 'addEventListener' is being called on
   * @example
   * myObject.addEventListener('eventName', (eventParameter1, eventParameter2) => {
   *  ...
   * });
   */
  addEventListener(type, fn, options) {
    const once = options && options.once;
    
    return this._on(type, fn, {
      external: true,
      once,
    });
  }

  /**
   * Remove a handler of the given event name and namespace (if given) or with a function reference
   * @param {string|number} [type] - The name of the event to remove the handler of with an optional namespace.
   * @param {function} [fn] - The handler associated with this event to be removed. If fn is undefined, all the handlers of the given event namespace will be removed. If you are not passing in this parameter then a namespace must be used with the event name.
   * @returns {object} Returns the object that 'removeEventListener' is being called on
   * @example
   * myObject.removeEventListener('eventName.namespace');
   * myObject.removeEventListener('eventName', fn);
   */
  removeEventListener(type, fn) {
    return this._off(type, fn);
  }

  /**
   * Calls the handlers of the event name with given data
   * @param {string|number} type - event name of which the handlers will be called.
   * @param {*} [data] - data that will be passed to the handlers. If data is an array, it will be spread and then passed to the handlers
   * @returns {object} Returns the object that 'trigger' is being called on
   * @example
   * myObject.trigger('eventName');
   * myObject.trigger('eventName', [eventParameter1, eventParameter2]);
   */
  trigger(type, data, action = '') {
    const eventListeners = this._listeners[type];
    if (!eventListeners || eventListeners.length === 0) {
      return;
    }
    let externalListeners;
    // Making copy of the listeners array in case if it will be modified
    // during dispatch.
    for (const { fn, external, once } of eventListeners.slice(0)) {
      if (once) {
        this._off(type, fn)
      }
      if (external) {
        (externalListeners ||= []).push(fn)
        continue;
      }
      fn(data, action)
    }
    // Dispatch any "external" listeners *after* the internal ones, to give the
    // viewer components time to handle events and update their state first.
    const listenersToRemove = externalListeners;
    if (externalListeners) {
      for (const fn of externalListeners) {
        fn(data);
      }
      externalListeners = null;
    }
    return listenersToRemove;
  }

  /**
   * @ignore
   */
  _on(type, fn, options = null) {
    const eventListeners = (this._listeners[type] ||= []);
    const listener = {
      fn,
      external: options?.external === true,
      once: options?.once === true,
    };
    eventListeners.push(listener);
    return listener;
  }

  /**
   * @ignore
   */
  _off(type, fn) {
    const eventListeners = this._listeners[type];
    if (!eventListeners) {
      return;
    }
    if (!fn) {
      const listenersToRemove = eventListeners;
      eventListeners.length = 0;
      return listenersToRemove;
    }
    for (let i = 0, ii = eventListeners.length; i < ii; i++) {
      if (eventListeners[i].fn === fn) {
        return eventListeners.splice(i, 1);
      }
    }
  }
}

export default EventHandler;
