/**
 * A utility class that represents a point with an x and y coordinate.
 * @memberof Core.Math
 */
class Point {

  /** Gets a point at (0, 0). */
  static get Origin() {
    return new Point(0, 0);
  }

  /**
   * Creates a new Point with an x and y coordinate.
   * @param {number} x - The x-coordinate
   * @param {number} y - The y-coordinate
   * @property {number} x - The x-coordinate
   * @property {number} y - The y-coordinate
   */
  constructor(x, y) {
    this.x = x;
    this.y = y;
  }

  /**
   * Adds two vectors together.
   * @param {Core.Math.Point} vec - A delta vector
   * @return {Core.Math.Point} The combined vector representing a new direction
   */
  add(vec) {
    return new Point(this.x + vec.x, this.y + vec.y);
  }

  /**
   * Gets the angle between this point and another relative to the origin.
   * @param {Core.Math.Point} vec - The other point
   * @param {boolean} [inRadians=false] - Whether the result should be in radians. Default: false
   * @returns {number} The angle between these two points. Returns radians if inRadians is true
   */
  angle(vec, inRadians = false) {
    const dx = vec.x - this.x;
    const dy = vec.y - this.y;
    const angle = Math.atan2(dy, dx);
    return inRadians ? angle : angle * (180 / Math.PI); // 转换为度
  }

  /**
   * Clones this point.
   * @returns Returns a new instance of the same point.
   */
  clone() {
    return new Point(this.x, this.y);
  }

  /**
   * Gets the cross product of this vector and another. This operation is non-commutative.
   * @param {Core.Math.Point} vec - The other vector
   * @returns {*} A cross product object with X, Y, Z properties
   */
  cross(vec) {
    return {
      X: 0, // 由于是二维点，X 组件为 0
      Y: 0, // 由于是二维点，Y 组件为 0
      Z: this.x * vec.y - this.y * vec.x // 计算叉积
    };
  }

  /**
   * Gets the distance between this point and another.
   * @param {Core.Math.Point} vec - The other vector
   * @returns {number} The distance between these two points
   */
  distance(vec) {
    const dx = vec.x - this.x;
    const dy = vec.y - this.y;
    return Math.sqrt(dx * dx + dy * dy); // 计算欧几里得距离
  }

  /**
   * Gets the dot product of this vector and another. This operation is commutative.
   * @param {Core.Math.Point} vec - The other vector
   * @returns {number} The dot product 
   */
  dot(vec) {
    return this.x * vec.x + this.y * vec.y; // 计算点积
  }

  /**
   * Compares this point to another. Points are considered the same if they are within 0.00001 units of each other.
   * @param {Core.Math.Point} vec - The other point
   * @returns Returns true if points are 0.00001 units of each other. Otherwise false.
   */
  equals(vec) {
    return this.distance(vec) < 0.00001; // 判断距离是否小于阈值
  }

  /**
   * Returns the X value of this point.
   * @returns {number} The X value
   */
  getX() {
    return this.x;
  }

  /**
   * Returns the Y value of this point.
   * @returns {number} The Y value
   */
  getY() {
    return this.y;
  }

  /**
   * Checks whether this vector is parallel to another vector.
   * @param {Core.Math.Point} vec - The other point
   * @returns Returns true if both vectors are parallel (same slope). Otherwise false.
   */
  isParallelTo(vec) {
    // 检查斜率是否相同（避免除以零的情况）
    return this.x * vec.y === this.y * vec.x;
  }

  /**
   * Normalizes the vector to get a unit vector of length 1.
   * @returns {Core.Math.Point} The unit vector
   */
  normalize() {
    const length = Math.sqrt(this.x * this.x + this.y * this.y);
    return length > 0 ? new Point(this.x / length, this.y / length) : new Point(0, 0); // 返回单位向量
  }

  /**
   * Gets a perpendicular vector from this one.
   * @returns {Core.Math.Point} The perpendicular vector
   */
  perpendicular() {
    return new Point(-this.y, this.x); // 返回垂直向量
  }

  /**
   * Reverses the direction of this vector.
   * @returns {Core.Math.Point} The reverse vector
   */
  reverse() {
    return new Point(-this.x, -this.y); // 返回反向向量
  }

  /**
   * Rotates this point around another point clockwise. By default, this will rotate around the origin.
   * @param {number} angle - The angle to rotate by
   * @param {boolean} [inRadians=false] - Whether the amount is in radians. Default: false
   * @param {Core.Math.Point} [withRespectTo] - The point to rotate around. Default: (0, 0)
   */
  rotate(angle, inRadians = false, withRespectTo = Point.Origin) {
    const angleRad = inRadians ? angle : angle * (Math.PI / 180);
    const cos = Math.cos(angleRad);
    const sin = Math.sin(angleRad);

    // 转换为以原点为中心的坐标系
    const translatedX = this.x - withRespectTo.x;
    const translatedY = this.y - withRespectTo.y;

    // 应用旋转变换
    const rotatedX = translatedX * cos - translatedY * sin + withRespectTo.x;
    const rotatedY = translatedX * sin + translatedY * cos + withRespectTo.y;

    return new Point(rotatedX, rotatedY);
  }

  /**
   * Scales the point by a scaling factor.
   * @param {number} s - The amount to scale by
   */
  scale(s) {
    return new Point(this.x * s, this.y * s); // 按比例缩放
  }

  /**
   * Subtracts two vectors from each other.
   * @param {Core.Math.Point} vec - A delta vector
   * @returns {Core.Math.Point} The resulting vector representing a new direction
   */
  subtract(vec) {
    return new Point(this.x - vec.x, this.y - vec.y); // 减去向量
  }

  /**
   * Serialize a Point object to a JSON object, which can be saved for further use. Calling 'JSON.stringify()' on a Point object will call this function automatically.
   * @returns The converted Point JSON object.
   */
  toJSON() {
    return { x: this.x, y: this.y }; // 转换为 JSON 对象
  }

  /**
   * Gets a matrix that represents this point. You can include additional rows to add to the final matrix.
   * @param {Array.<Core.Math.Point>} [vecs] - Additional rows to add.
   * @returns {Core.Math.Point} The matrix that represents this point.
   */
  toMatrix(vecs = []) {
    const matrix = [[this.x, this.y, 1]]; // 将点转化为矩阵形式
    vecs.forEach(vec => {
      matrix.push([vec.x, vec.y, 1]); // 添加更多行
    });
    return matrix; // 返回表示的矩阵
  }

  /**
   * Applies a transformation matrix to this point.
   * @param {Core.Math.Matrix} transformMatrix - The transformation matrix.
   */
  transform(transformMatrix) {
    const x = transformMatrix[0][0] * this.x + transformMatrix[0][1] * this.y + transformMatrix[0][2];
    const y = transformMatrix[1][0] * this.x + transformMatrix[1][1] * this.y + transformMatrix[1][2];
    return new Point(x, y); // 应用变换矩阵并返回新的点
  }

  /**
   * Translates this point by the specified delta values.
   * @param {number} dx - the unit in the x-axis to translate by
   * @param {number} dy - the unit in the y-axis to translate by
   */
  translate(dx, dy) {
    return new Point(this.x + dx, this.y + dy); // 平移点
  }
}

export default Point;
