/**
 * Represents a utility class used to manipulate rectangle objects.
 * @memberof Core.Math
 */
class Rect {

  /**
   * Create a new Rect with the x and y coordinates of the upper-left and lower right points.
   * @param {number} x1 - the x coordinate of the upper-left point
   * @param {number} y1 - the y coordinate of the upper-left point
   * @param {number} x2 - the x coordinate of the lower-right point
   * @param {number} y2 - the y coordinate of the lower-right point
   * 
   * @property {number} x1 - the x coordinate of the upper-left point
   * @property {number} y1 - the y coordinate of the upper-left point
   * @property {number} x2 - the x coordinate of the lower-right point
   * @property {number} y2 - the y coordinate of the lower-right point
   * 
   * @example
(x1,y1)------|
  |          |
  |-------(x2,y2)
   */
  constructor(x1, y1, x2, y2) {
    this.x1 = x1;
    this.y1 = y1;
    this.x2 = x2;
    this.y2 = y2;
  }

  /**
   * Creates a Rect based on the combined bounding box of the provided rects.
   * @param {Array.<Rect>} rectArray - The set of rects to create the rect from
   * @returns {Core.Math.Rect} A rect at the designated position of the defined size
   */
  static combineRects(rectArray) {
    if (rectArray.length === 0) return null;

    let x1 = Math.min(...rectArray.map(rect => rect.x1));
    let y1 = Math.min(...rectArray.map(rect => rect.y1));
    let x2 = Math.max(...rectArray.map(rect => rect.x2));
    let y2 = Math.max(...rectArray.map(rect => rect.y2));

    return new Rect(x1, y1, x2, y2);
  }

  /**
   * A static method that determines whether the first rect is contains the second rect.
   * @param {Core.Math.Rect} rect - The first rect (container)
   * @param {Core.Math.Rect} container - The second rect
   * @returns {boolean} True if the first rect contains the second rect. Otherwise false
   */
  static contains(rect, container) {
    return rect.x1 >= container.x1 &&
      rect.y1 >= container.y1 &&
      rect.x2 <= container.x2 &&
      rect.y2 <= container.y2;
  }

  /**
   * Creates a Rect based on the combined bounding box of the provided annotations.
   * @param {Array.<Annotations.Annotation>} annotations - The set of annotations to create the rect from
   * @returns {Core.Math.Rect} A rect at the designated position of the defined size
   */
  static createEncapsulatingBoundingBox(annotations) {
    if (annotations.length === 0) return null;

    // Assume annotations have properties x1, y1, x2, y2
    let x1 = Math.min(...annotations.map(annotation => annotation.x1));
    let y1 = Math.min(...annotations.map(annotation => annotation.y1));
    let x2 = Math.max(...annotations.map(annotation => annotation.x2));
    let y2 = Math.max(...annotations.map(annotation => annotation.y2));

    return new Rect(x1, y1, x2, y2);
  }

  /**
   * Creates a Rect using position and dimensions.
   * @param {number} x - The X coordinate
   * @param {number} y - The Y coordinate
   * @param {number} width - The width of the rect
   * @param {number} height - The height of the rect
   * @returns {Core.Math.Rect} A rect at the designated position of the defined size
   */
  static createFromDimensions(x, y, width, height) {
    return new Rect(x, y, x + width, y + height);
  }

  /**
   * A static method that determines whether the two rects provided intersect each other.
   * @param {Core.Math.Rect} rect1 - The first rect
   * @param {Core.Math.Rect} rect2 - The second rect
   * @returns {boolean} True if the rects intersect. Otherwise, false.
   */
  static intersects(rect1, rect2) {
    return !(rect2.x1 > rect1.x2 ||
      rect2.x2 < rect1.x1 ||
      rect2.y1 > rect1.y2 ||
      rect2.y2 < rect1.y1);
  }

  /**
   * A method to determine whether this rect contains another rect. It is not considered as containment if another rect intersects this one.
   * @param {Core.Math.Rect} otherRect - The other rect
   * @returns {boolean} True if this rect contains 'otherRect'. Otherwise, false.
   */
  contains(otherRect) {
    return this.x1 < otherRect.x1 && this.y1 < otherRect.y1 &&
      this.x2 > otherRect.x2 && this.y2 > otherRect.y2 &&
      !this.intersects(otherRect);
  }

  /**
   * Normalizes the rect to ensure point (x1, y1) is the bottom left point as expected by xfdf output
   */
  exportNormalize() {
    if (this.y1 > this.y2) {
      [this.y1, this.y2] = [this.y2, this.y1];
    }
  }

  /**
   * Best fit this rect into another larger rect
   * @param {Core.Math.Rect} rect 
   */
  fitTo(rect) {
    const scaleX = (rect.x2 - rect.x1) / this.getWidth();
    const scaleY = (rect.y2 - rect.y1) / this.getHeight();
    const scale = Math.min(scaleX, scaleY);
    const newWidth = this.getWidth() * scale;
    const newHeight = this.getHeight() * scale;

    const centerX = rect.x1 + (rect.x2 - rect.x1) / 2;
    const centerY = rect.y1 + (rect.y2 - rect.y1) / 2;

    this.x1 = centerX - newWidth / 2;
    this.y1 = centerY - newHeight / 2;
    this.x2 = centerX + newWidth / 2;
    this.y2 = centerY + newHeight / 2;
  }

  /**
   * Returns the most bottom Y value.
   * @returns {number}
   */
  getBottom() {
    return this.y2;
  }

  /**
   * Gets a [Point]{@link Core.Math.Point} that is at the bottom-left of the rect.
   * @returns {Core.Math.Point} A point at the bottom-left of the rect
   */
  getBottomLeft() {
    return new Point(this.x1, this.y2);
  }

  /**
   * Gets a [Point]{@link Core.Math.Point} that is at the top-right of the rect.
   * @returns {Core.Math.Point} A point at the top-right of the rect
   */
  getBottomRight() {
    return new Point(this.x2, this.y2);
  }

  /**
   * Gets a [Point]{@link Core.Math.Point} that is at the center of the rect.
   * @returns {Core.Math.Point} A point at the center of the rect
   */
  getCenter() {
    return new Point((this.x1 + this.x2) / 2, (this.y1 + this.y2) / 2);
  }

  /**
   * Gets the height of the rect
   * @returns {number} The height of the rect
   */
  getHeight() {
    return this.y2 - this.y1;
  }

  /**
   * Returns the most left X value.
   * @returns {number}
   */
  getLeft() {
    return this.x1;
  }

  /**
   * Returns the most right X value.
   * @returns {number}
   */
  getRight() {
    return this.x2;
  }

  /**
   * Returns the most top Y value.
   * @returns {number}
   */
  getTop() {
    return this.y1;
  }

  /**
   * Gets a [Point]{@link Core.Math.Point} that is at the top-left of the rect.
   * @returns {Core.Math.Point} A point at the top-left of the rect
   */
  getTopLeft() {
    return new Point(this.x1, this.y1);
  }

  /**
   * Gets a [Point]{@link Core.Math.Point} that is at the top-right of the rect.
   * @returns {Core.Math.Point} A point at the top-right of the rect
   */
  getTopRight() {
    return new Point(this.x2, this.y1);
  }

  /**
   * Gets the width of the rect
   * @returns {number}
   */
  getWidth() {
    return this.x2 - this.x1;
  }

  /**
   * A method to determine whether this rect overlaps with another. If this rect contains another, this is still considered true.
   * @param {Core.Math.Rect} otherRect - The other rect
   * @returns {boolean} True if the rects intersect. Otherwise, false.
   */
  intersects(otherRect) {
    return !(otherRect.x1 > this.x2 ||
      otherRect.x2 < this.x1 ||
      otherRect.y1 > this.y2 ||
      otherRect.y2 < this.y1);
  }

  /**
   * Joins two Rects together into one Rect that contains both.
   * @param {Core.Math.Rect} rect 
   * @returns {Core.Math.Rect}
   */
  join(rect) {
    const x1 = Math.min(this.x1, rect.x1);
    const y1 = Math.min(this.y1, rect.y1);
    const x2 = Math.max(this.x2, rect.x2);
    const y2 = Math.max(this.y2, rect.y2);
    return new Rect(x1, y1, x2, y2);
  }

  /** Normalizes the rect to ensure point (x1, y1) is the upper left point and width and height are non-negative. */
  normalize() {
    // Ensure proper order of coordinates
    if (this.x1 > this.x2) [this.x1, this.x2] = [this.x2, this.x1];
    if (this.y1 > this.y2) [this.y1, this.y2] = [this.y2, this.y1];
  }

  /**
   * Returns a [Quad]{@link Core.Math.Quad} that represents this rect.
   * @returns {Core.Math.Quad} A quad that represents this rect
   */
  toQuad() {
    return {
      topLeft: this.getTopLeft(),
      topRight: this.getTopRight(),
      bottomRight: this.getBottomRight(),
      bottomLeft: this.getBottomLeft()
    };
  }

  /**
   * Applies a transformation matrix to the rect which will adjust the points.
   * @param {Core.Math.Matrix} matrix - The transformation matrix to apply
   */
  transform(matrix) {
    // Apply transformation logic using matrix
    // Placeholder: Actual transformation logic to be implemented
  }

  /**
   * Translates the rect
   * @param {number} x - The amount to translate in the x direction
   * @param {number} y - The amount to translate in the y direction
   */
  translate(x, y) {
    this.x1 += x;
    this.x2 += x;
    this.y1 += y;
    this.y2 += y;
  }
}

export default Rect;
